#
# Cookbook:: gitlab-ci-runner
# Library:: runner
#
# Copyright:: 2020, Ben Hughes <bmhughes@bmhughes.co.uk>, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'json'

require_relative '_common'
require_relative 'api'
require_relative 'resource'
require_relative 'toml'

module GitlabCiRunner
  module Cookbook
    # Helper methods for configuring gitlab-runner via chef
    module RunnerHelpers
      include GitlabCiRunner::Cookbook::ApiHelpers
      include GitlabCiRunner::Cookbook::CommonHelpers
      include GitlabCiRunner::Cookbook::ResourceHelpers
      include GitlabCiRunner::Cookbook::TomlHelpers

      RUNNER_EXECUTOR_REGEXP ||= /(docker|parallels|virtualbox|ssh|machine|kubernetes)(_|-)/.freeze
      RUNNER_SERVER_OPTIONS ||= %i(description active tag_list run_untagged locked access_level maximum_timeout).freeze
      RUNNER_LOCAL_OPTIONS ||= %i(name url).freeze
      RUNNER_LOCAL_REMOVE ||= %w(registration_token).freeze
      RUNNER_LOCAL_PRESERVE ||= %w(name token).freeze

      # Update an in-place local configuration as the token option must be preserved so we can't
      # simply overwrite the existing configuration with the given one.
      def runner_update_local(config_index, options)
        if nil_or_empty?(config_resource_runners[config_index])
          log_chef(:debug, "No runners configuration present for index #{config_index}.")
          return config_resource_runners[config_index]
        end

        log_chef(:info, "Checking options for runner #{config_resource_runners[config_index]['name']}.")
        config_change_flag = false

        options.each do |option, value|
          next if nil_or_empty?(value)

          log_chef(:debug, "Testing option #{option}.")
          if config_resource_runners[config_index][option].eql?(value)
            log_chef(:debug, "Option #{option} matches. Current: #{config_resource_runners[config_index][option]}. Incoming: #{value}).")
            next
          end

          log_chef(:info, "Option #{option} has changed. Current: #{config_resource_runners[config_index][option]}. Incoming: #{value}.")
          config_resource_runners[config_index][option] = value
          config_change_flag = true
        end

        option_cleanup(config_resource_runners[config_index], options)

        if config_change_flag
          log_chef(:info, "Runner #{config_resource_runners[config_index]['name']} options updated.")
          true
        else
          log_chef(:info, "Runner #{config_resource_runners[config_index]['name']} unmodified, options match.")
          false
        end
      end

      # Convert options hash from 'exec-option' keys to a nested Hash
      def runner_options_to_h(options)
        options_new = options.dup

        options.each_key do |option|
          next unless runner?(option)

          a = runner_option_split(option)
          options_new[a[0]] ||= {}
          options_new[a[0]][a[1]] = options_new.delete(option)
        end

        options_new
      end

      def runner_registered?(name, host, http_options)
        return false unless runner_configured?(name, host)

        token, url = runner_get_token(name)

        log_chef(:info, "Checking if runner #{name} is registered with #{host}.")
        if api_runner_valid?(url, http_options, token)
          log_chef(:info, "Runner #{name} is registered with #{host}.")
          true
        else
          log_chef(:info, "Runner #{name} is NOT registered with #{host}.")
          false
        end
      end

      def runner_configured?(name, host)
        token, url = runner_get_token(name)

        if !token.nil? && url.eql?(host)
          log_chef(:info, "Runner #{name} is configured for host #{host} with token #{token}.")
          true
        else
          log_chef(:info, "Runner #{name} NOT configured for host #{host}.")
          false
        end
      end

      def runner_register(host, http_options, registration_token, options)
        raise_error(type: ArgumentError, message: 'Registration requires a registration token to be passed') if nil_or_empty?(registration_token)

        options['token'] = registration_token

        response = api_runner_register(
          host,
          http_options,
          option_join(options)
        )

        if api_response_is?(response, Net::HTTPCreated)
          response = JSON.parse(response.body)
          log_chef(:info, "Runner #{response['id']} created sucessfully on #{host} with token #{response['token']}.")

          [response['id'], response['token']]
        else
          raise_error(message: "#{response.code} Error when registering: #{response}")
        end
      end

      def runner_unregister!(host, http_options, token)
        response = api_runner_remove!(host, http_options, token)

        if api_response_is?(response, Net::HTTPNoContent)
          log_chef(:info, "Runner with token #{token} successfully unregistered from #{host}.")
          true
        else
          log_chef(:warn, "Error unregistering runner with token #{token} from #{host}.")
          false
        end
      end

      def runner_unconfigure!(token, url)
        log_chef(:info, "Unconfiguring runner with token #{token} and url #{url}.")
        config_resource_runners.delete_if { |r| r['token'].eql?(token) && r['url'].eql?(url) }
      end

      def runner_update_server(host, http_options, token, id, options)
        if api_response_is?(api_runner_update!(host, http_options, token, id, option_join(options)), Net::HTTPSuccess)
          log_chef(:info, "Runner #{id} updated successfully on #{host}.")
          return true
        end

        log_chef(:warn, "Error updating runner #{id} on #{host}.")

        false
      end

      def runner_update_needed?(host, http_options, token, id, options)
        log_chef(:info, "Checking runner update required on server #{host} for runner #{id}.")

        response = api_runner_get(host, http_options, token, id)
        raise_error(message: "Error #{response.code} checking runner update required.") unless api_response_is?(response, Net::HTTPSuccess)

        log_chef(:debug, "Runner found, response code: #{response.code}.")

        if runner_options_diff(JSON.parse(response.body), options).empty?
          log_chef(:info, "Runner #{id} does not require update.")
          false
        else
          log_chef(:info, "Runner #{id} requires update.")
          true
        end
      end

      def runner_options_diff(server, local)
        log_chef(:debug, "Server runner: #{server}")
        log_chef(:debug, "Local options: #{local}")

        local.reject do |option, value|
          log_chef(:info, "Testing option #{option}. Local: #{value}. Remote: #{server[option]}.")
          log_chef(:info, "Option #{option} differs. Local: #{value}. Server: #{server[option]}.") unless safe_compare(server[option], value)
          safe_compare(server[option], value)
        end
      end

      def runner_get_token(name, id = nil)
        config_resource_runners.each do |r|
          if runner_split_id_name(r['name']).last.eql?(name) && (runner_split_id_name(r['name']).first.eql?(id) || id.nil?)
            log_chef(:info, "Found runner #{name} with token #{r['token']} registered at url #{r['url']}.")
            return [r['token'], r['url']]
          end
        end

        log_chef(:debug, "Runner #{name} not found.")

        nil
      end

      def runner_get_index(url, token)
        i = config_resource_runners.find_index { |r| r['url'].eql?(url) && r['token'].eql?(token) }

        raise_error(message: "Unable to find index for token #{token} and url #{url}.") if i.nil?
        log_chef(:info, "Found runner configuration for token #{token} and url #{url} at index #{i}.")

        i
      end

      def runner_get_id(runner_name)
        config_resource_runners.each do |r|
          id, name = runner_split_id_name(r['name'])

          if name.eql?(runner_name)
            log_chef(:info, "Found ID #{id} for runner #{runner_name}.")
            return id
          end
        end

        log_chef(:warn, "No ID found for runner #{runner_name}.")

        nil
      end

      private

      # Remove options that have been deleted
      def option_cleanup(config, options)
        removed_options = config.keys.reject { |k| options.key?(k) || RUNNER_LOCAL_PRESERVE.include?(k) }
        if removed_options.empty?
          log_chef(:debug, 'Removed options empty, nothing to cleanup.')
        else
          log_chef(:debug, "Deleting removed options #{removed_options.join(', ')}.")
          config.reject! { |k, _| removed_options.include?(k) || RUNNER_LOCAL_REMOVE.include?(k) }
        end
      end

      # Join any Array values in Hash
      def option_join(options)
        return options if options.empty?

        options.map { |option, value| [option, value.respond_to?(:join) ? value.join(',') : value] }.to_h
      end

      # Option matches a valid runner executor
      def runner?(option)
        option.match(RUNNER_EXECUTOR_REGEXP)
      end

      # Split 'executor-option' String into Array of ['executor', 'option']
      def runner_option_split(option)
        option.split(RUNNER_EXECUTOR_REGEXP).delete_if { |string| string.empty? || %w(- _).include?(string) }
      end

      # Split runner configuration file id:name String into Array
      def runner_split_id_name(name)
        return [nil, name] if name.split(':').count.eql?(1)

        result = [name.split(':').first.to_i, name.split(':')[1].strip]
        log_chef(:debug, "Split result: #{result}.")

        result
      end
    end
  end
end
