#
# Cookbook:: gitlab-ci-runner
# Resource:: runner
#
# Copyright:: 2020, Ben Hughes <bmhughes@bmhughes.co.uk>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

include GitlabCiRunner::Cookbook::Helpers

property :config_file, String,
          default: lazy { default_gitlab_runner_config_file },
          description: 'Configuration file location'

property :cookbook, String,
          default: 'gitlab-ci-runner',
          description: 'Cookbook to source configuration file template from'

property :template, String,
          default: 'gitlab-runner/config.toml.erb',
          description: 'Template to use to generate the configuration file'

property :owner, String,
          default: lazy { default_gitlab_runner_user },
          description: 'Owner of the generated configuration file'

property :group, String,
          default: lazy { default_gitlab_runner_group },
          description: 'Group of the generated configuration file'

property :mode, String,
          default: '0600',
          description: 'Filemode of the generated configuration file'

property :sensitive, [true, false],
          default: true

property :url, String,
          description: 'Gitlab Server URL'

property :registration_token, String,
          description: 'Server runner registation token'

property :administration_token, String,
          description: 'Server user administration token for updating runner server-side options'

property :http_options, Hash,
          default: lazy { default_gitlab_runner_http_options },
          description: 'Set net/http options for API requests',
          coerce: proc { |p| p.merge(default_gitlab_runner_http_options) }

property :description, String,
          default: lazy { name },
          description: 'Server - Runner description'

property :active, [true, false],
          description: 'Server - Runner active flag'

property :tag_list, Array,
          default: [],
          description: 'Server - Runner tag list'

property :run_untagged, [true, false],
          description: 'Server - Runner run untagged option'

property :locked, [true, false],
          description: 'Server - Runner locked option'

property :access_level, String,
          equal_to: %w(not_protected ref_protected),
          description: 'Server - Runner access level'

property :maximum_timeout, Integer,
          description: 'Server - Runner maximum timeout value'

property :options, Hash,
          default: {},
          description: 'Local - Runner options'

action_class do
  include GitlabCiRunner::Cookbook::ResourceHelpers
  include GitlabCiRunner::Cookbook::RunnerHelpers

  def server_setting(option)
    property_is_set?(option) ? new_resource.send(option).dup : new_resource.options.fetch(option.to_s, nil)
  end

  def server_options
    compact_hash(GitlabCiRunner::Cookbook::RunnerHelpers::RUNNER_SERVER_OPTIONS.map { |p| [p.to_s, server_setting(p)] }.to_h)
  end

  def local_options
    local = GitlabCiRunner::Cookbook::RunnerHelpers::RUNNER_LOCAL_OPTIONS.map { |p| [p.to_s, new_resource.send(p).dup] }.to_h
    local.merge!(runner_options_to_h(new_resource.options.dup))
    local.delete_if { |option, _| GitlabCiRunner::Cookbook::RunnerHelpers::RUNNER_SERVER_OPTIONS.include?(option) }
  end

  def validate_options
    raise ArgumentError, 'Keys of the Hash provided to the options property MUST be of type String' unless hash_keys_string?(new_resource.options)
  end
end

action :register do
  validate_options
  config_resource_init

  unless runner_registered?(new_resource.name, local_options['url'], new_resource.http_options)
    id, token = nil
    # Register with Gitlab server
    converge_by("Register gitlab-runner #{new_resource.name} with server #{new_resource.url}.") do
      id, token = runner_register(new_resource.url,
                                  new_resource.http_options,
                                  new_resource.registration_token,
                                  server_options)

      log "Runner #{new_resource.name} registered with #{new_resource.url} as ID: #{id}."
    end

    # Local configuration
    converge_by("Add gitlab runner #{new_resource.name} to local configuration.") do
      new_runner = local_options
      new_runner['name'].prepend("#{id}: ")
      new_runner['token'] = token

      new_runner = compact_hash(new_runner)
      log_chef(:info, "Adding new runner #{new_runner['name']} with config #{new_runner} to configuration file #{new_resource.config_file}.")
      config_resource_runners.push(new_runner)
    end
  end
end

action :update do
  validate_options
  config_resource_init

  if runner_registered?(new_resource.name, local_options['url'], new_resource.http_options)
    token, url = runner_get_token(new_resource.name)
    runner_index = runner_get_index(url, token)

    # Local configuration
    converge_by("Update runner #{new_resource.name} configuration") {} if runner_update_local(runner_index, local_options.delete_if { |key, _| key.eql?('name') })

    # Server configuration
    id = runner_get_id(new_resource.name)

    if property_is_set?(:administration_token)
      if runner_update_needed?(url, new_resource.http_options, new_resource.administration_token, id, server_options)
        converge_by("Updating runner #{id} on server #{url}.") do
          runner_update_server(url, new_resource.http_options, new_resource.administration_token, id, server_options)
        end
      end
    else
      log_chef(:warn, "\n     Unable to check/update server options for Runner: #{new_resource.name} | ID: #{id} as an administration token has not been specified.")
    end
  else
    raise "Unable to update unregistered runner #{new_resource.name} for url #{local_options['url']}."
  end
end

action :unregister do
  config_resource_init

  token, url = runner_get_token(new_resource.name)
  return unless runner_configured?(new_resource.name, url)

  if runner_registered?(new_resource.name, url, new_resource.http_options)
    converge_by("Unregister runner #{new_resource.name} from server #{url}.") { runner_unregister!(url, new_resource.http_options, token) }
  end

  converge_by("Remove runner #{new_resource.name} from local configuration.") { runner_unconfigure!(token, url) } if runner_configured?(new_resource.name, url)
end
